# Python
from datetime import datetime, date
import decimal
import json
from typing import Optional, Union, Callable, Tuple, overload

# This App
from .exceptions import InputError


@overload
def string_validation(text: Optional[str], max_length: int, obj: str, empty: bool) -> Optional[str]: # pragma: no coverage
    ...

@overload
def string_validation(text: Optional[str], max_length: int, obj: str) -> str: # pragma: no coverage
    ...

def string_validation(text: Optional[str], max_length: int, obj: str, empty: bool | None=False) :
    try:
        if empty is False and (text is None or len(text) == 0):
            raise InputError(obj=obj, message=' cannot be empty!')
        elif empty is True and (text is None or len(text) == 0):
            return None
        elif text is not None and len(text) > max_length:
            raise InputError(obj=obj, message=' cannot be more than ' + str(max_length) + ' characters!')
        else:
            return text
    except TypeError: # pragma: no coverage
        raise InputError(obj=obj, message=' is either missing or not a string')  # noqa: B904


@overload
def choice_validation(text: Optional[str], choices: Union[Tuple, list], obj: str, null: bool) -> Optional[str]: # pragma: no coverage
    ...

@overload
def choice_validation(text: Optional[str], choices: Union[Tuple, list], obj: str) -> str: # pragma: no coverage
    ...

def choice_validation(text: Optional[str], choices: Union[Tuple, list], obj: str, null: bool | None = False):
    try:
        if null is True and (text is None or len(text) == 0):
            return None
        elif null is False and (text is None or len(text) == 0):
            raise InputError(obj=obj, message=' has to be selected!')
        elif ((isinstance(choices, list) or type(choices) is tuple) and text in choices) or text in dict(choices):
            return text
        else:
            raise InputError(obj=obj, message=' cannot have this value!')
    except ValueError:
        raise InputError(obj=obj, message=' cannot have this value!') from None




@overload
def integer_validation(number: Optional[Union[str, int]], low: int, high: int, obj: str, empty: bool) -> Optional[int]:
    ...

@overload
def integer_validation(number: Optional[Union[str, int]], low: int, high: int, obj: str) -> int:
    ...

def integer_validation(number: Optional[Union[str, int]], low: int, high: int, obj: str, empty: bool | None = False):
    if empty and (number is None or (type(number) is str and len(number) == 0)):
        return None
    elif empty is False and (number is None or number == ""):
        raise InputError(obj=obj, message='has to be an integer!')
    else:
        try:
            n = int(number) # type: ignore
            if low <= n <= high:
                return n
            else:
                if n < low:
                    message = 'has to be equal or bigger than ' + str(low)
                else:
                    message = 'has to be equal or smaller than ' + str(high)
                raise InputError(obj=obj, message=message)
        except ValueError:
            raise InputError(obj=obj, message='has to be an integer!')  # noqa: B904


def boolean_validation(boo: Optional[Union[str, bool]], obj: str) -> bool:
    if boo is True or boo is False:
        return boo
    elif boo == 'true' or boo == 'True':
        return True
    elif boo == 'false' or boo == 'False':
        return False
    else:
        raise InputError(obj=obj, message=' has to be either true or false!')



@overload
def date_validation(target_date: str | None, obj: str, empty: bool, min_date: str | date | None=None) -> Optional[date]:
    ...


@overload
def date_validation(target_date: str | None, obj: str, empty: None=None, min_date: str | date | None=None) -> date:
    ...
    

def date_validation(target_date: str | None, obj: str, empty: bool | None = False, min_date : str | date | None =None):
    if empty and (target_date is None or (type(target_date) == str and len(target_date) == 0)):
        return None
    elif empty is False and (target_date is None or len(target_date) == 0):
        raise InputError(obj=obj, message=' cannot be empty!')
    else:
        try:
            formatted_date = datetime.strptime(target_date, "%Y-%m-%d").date() # type: ignore
            if min_date is None:
                return formatted_date
            else:
                formatted_min_date = datetime.strptime(min_date, "%Y-%m-%d").date() if type(min_date) == str else min_date
                if formatted_date >= formatted_min_date: # type: ignore
                    return formatted_date
                else:
                    raise InputError(obj=obj, message=' cannot be earlier than ' + str(min_date))
        except ValueError:
            raise InputError(obj=obj, message=' is not a valid date!')  # noqa: B904



@overload
def date_time_validation(target_date_time: str | None, obj: str, empty: bool, min_date: str | datetime | None=None) -> Optional[datetime]: # pragma: no coverage
    ...


@overload
def date_time_validation(target_date_time: str | None, obj: str, empty: None=None, min_date: str | datetime | None=None) -> datetime: # pragma: no coverage
    ...
    

def date_time_validation(target_date_time: str | None, obj: str, empty: bool | None = False, min_date : str | datetime | None =None):
    if empty and (target_date_time is None or (isinstance(target_date_time, str) and len(target_date_time) == 0)):
        return None
    elif empty is False and (target_date_time is None or len(target_date_time) == 0):
        raise InputError(obj=obj, message=' cannot be empty!')
    else:
        try:
            formatted_date = datetime.fromisoformat(target_date_time or "")
            if min_date is None:
                return formatted_date
            else:
                formatted_min_date_time = datetime.fromisoformat(min_date) if isinstance(min_date, str) else min_date
                if formatted_date >= formatted_min_date_time: # type: ignore
                    return formatted_date
                else:
                    raise InputError(obj=obj, message=' cannot be earlier than ' + formatted_min_date_time.isoformat())
        except ValueError:
            raise InputError(obj=obj, message=' is not a valid date!')  # noqa: B904
        except TypeError: # pragma: no coverage
            raise InputError(obj=obj, message=' is not a valid date!')  # noqa: B904



def decimal_validation(number, low, high, decimal_count, obj, round_decimal=True, empty=False, more_message=None):
    try:
        dec = decimal.Decimal(number)
        dec_places = dec.as_tuple().exponent
        if low <= dec <= high:
            if dec_places <= decimal_count:
                return dec
            else:
                if round_decimal:
                    return dec.quantize(decimal.Decimal(10) ** (decimal_count * -1))
                else:
                    raise InputError(obj=obj, message=' cannot have more than ' + decimal_count + ' places!')
        else:
            if dec < low:
                message = ' cannot be smaller than ' + str(low)
            else:
                if more_message is None:
                    message = ' cannot be larger than ' + str(high)
                else:
                    message = more_message
            raise InputError(obj=obj, message=message)
    except decimal.InvalidOperation:
        if empty and len(number) == 0:
            return decimal.Decimal(0)
        else:
            raise InputError(obj=obj, message=' is not a valid number!')  # noqa: B904
    except TypeError:
        raise InputError(obj=obj, message=' is not a valid number!')  # noqa: B904


def json_validation(j, obj, empty=False) -> dict:
    try:
        if len(j) == 0 and empty:
            return j
        elif len(j) == 0 and not empty:
            raise InputError(obj=obj, message=' cannot be empty!')

        return json.loads(j)

    except json.decoder.JSONDecodeError:
        raise InputError(obj=obj, message=' is not valid!')  # noqa: B904


def list_validation(raw_list: Union[list, str], obj: str, nature: Callable, empty: bool = False) -> list:
    converted_list = list(raw_list)
    if empty and len(converted_list) == 0:
        return []
    elif empty is False and len(converted_list) == 0:
        raise InputError(obj=obj, message=' cannot be empty!')
    else:
        temp_list = []
        for c in converted_list:
            if nature.__name__ != 'all' and nature.__name__ != 'any':
                temp_list.append(nature(c))
            else:
                temp_list.append(c)
        return temp_list


# Django
from django.test import TestCase

# Foundation
from django_carpet.xml_element import XMLElement

class CarpetXMLElementTest(TestCase):

    def test_xml_element(self):
        sample_trimble_trace_data = """<env:Envelope
            xmlns:env='http://schemas.xmlsoap.org/soap/envelope/'>
            <env:Header></env:Header>
            <env:Body>
                <ns2:pollTracesResponse
                    xmlns:ns2="http://fleetworks.acunia.com/fleet/service"
                    xmlns:ns3="http://fleetworks.acunia.com/fleet/fault">
                    <return>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T19:39:10.657Z</time>
                            <coordinate>
                                <latitude>52.42955</latitude>
                                <longitude>4.80397</longitude>
                            </coordinate>
                            <mileage>158043690</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>5.292</value>
                            </property>
                            <property>
                                <key>AFRE</key>
                                <value>&lt;Report id="WAITING" version="2" timestamp="1455703892236"> &lt;Answer question="REASON">
                                    &lt;Value option="02" value="true"/>
                                    &lt;/Answer>
                                    &lt;/Report>
                                </value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T19:54:10.663Z</time>
                            <coordinate>
                                <latitude>52.42955</latitude>
                                <longitude>4.80396</longitude>
                            </coordinate>
                            <mileage>158043690</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>4.356</value>
                            </property>
                        </traces>
                        <traces>
                            <type>71</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T19:59:04.267Z</time>
                            <coordinate>
                                <latitude>52.42955</latitude>
                                <longitude>4.80400</longitude>
                            </coordinate>
                            <mileage>158043690</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.204</value>
                            </property>
                        </traces>
                        <traces>
                            <type>7</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T19:59:22.605Z</time>
                            <coordinate>
                                <latitude>52.42955</latitude>
                                <longitude>4.80401</longitude>
                            </coordinate>
                            <mileage>158043690</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.204</value>
                            </property>
                        </traces>
                        <traces>
                            <type>8</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T19:59:23.485Z</time>
                            <coordinate>
                                <latitude>52.42955</latitude>
                                <longitude>4.80401</longitude>
                            </coordinate>
                            <mileage>158043690</mileage>
                            <heading>178</heading>
                            <speed>4</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.204</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:04:23.493Z</time>
                            <coordinate>
                                <latitude>52.42926</latitude>
                                <longitude>4.80400</longitude>
                            </coordinate>
                            <mileage>158043725</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.168</value>
                            </property>
                        </traces>
                        <traces>
                            <type>196</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:04:30.867Z</time>
                            <coordinate>
                                <latitude>52.42926</latitude>
                                <longitude>4.80400</longitude>
                            </coordinate>
                            <mileage>158043725</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>PST</key>
                                <value>2020-10-05T19:04:25.511</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.168</value>
                            </property>
                            <property>
                                <key>PTRIG</key>
                                <value>TMT</value>
                            </property>
                            <property>
                                <key>XPID</key>
                                <value>43-BLK-9 (TAB)</value>
                            </property>
                        </traces>
                        <traces>
                            <type>55</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:05:16.632Z</time>
                            <coordinate>
                                <latitude>52.42925</latitude>
                                <longitude>4.80401</longitude>
                            </coordinate>
                            <mileage>158043725</mileage>
                            <heading>0</heading>
                            <speed>2</speed>
                            <property>
                                <key>STL</key>
                                <value>300000</value>
                            </property>
                            <property>
                                <key>STT</key>
                                <value>2020-10-05T19:59:43.510</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.168</value>
                            </property>
                            <property>
                                <key>STD</key>
                                <value>333122</value>
                            </property>
                        </traces>
                        <traces>
                            <type>13</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:08:08.939Z</time>
                            <coordinate>
                                <latitude>52.42639</latitude>
                                <longitude>4.80149</longitude>
                            </coordinate>
                            <mileage>158044485</mileage>
                            <heading>192</heading>
                            <speed>30</speed>
                            <property>
                                <key>ATY</key>
                                <value>UN</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601924647658-81</value>
                            </property>
                            <property>
                                <key>ALEN</key>
                                <value>3841270</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48115.6</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6283</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T19:04:07.659</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>DRLEN</key>
                                <value>525931</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.168</value>
                            </property>
                        </traces>
                        <traces>
                            <type>10</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:08:08.949Z</time>
                            <coordinate>
                                <latitude>52.42639</latitude>
                                <longitude>4.80149</longitude>
                            </coordinate>
                            <mileage>158044485</mileage>
                            <heading>192</heading>
                            <speed>30</speed>
                            <property>
                                <key>ATY</key>
                                <value>DR</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601928488930-80</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48115.6</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6284</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T20:08:08.929</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.168</value>
                            </property>
                        </traces>
                        <traces>
                            <type>167</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:09:04.502Z</time>
                            <coordinate>
                                <latitude>52.42525</latitude>
                                <longitude>4.80416</longitude>
                            </coordinate>
                            <mileage>158044920</mileage>
                            <heading>264</heading>
                            <speed>10</speed>
                            <property>
                                <key>FLP</key>
                                <value>99</value>
                            </property>
                            <property>
                                <key>FLC</key>
                                <value>10000</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.24</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:14:04.509Z</time>
                            <coordinate>
                                <latitude>52.42519</latitude>
                                <longitude>4.80459</longitude>
                            </coordinate>
                            <mileage>158044955</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>10</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:15:08.493Z</time>
                            <coordinate>
                                <latitude>52.42519</latitude>
                                <longitude>4.80458</longitude>
                            </coordinate>
                            <mileage>158044955</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ATY</key>
                                <value>UN</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601928908423-79</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48115.9</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6285</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T20:09:35.106</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>13</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:15:08.442Z</time>
                            <coordinate>
                                <latitude>52.42519</latitude>
                                <longitude>4.80458</longitude>
                            </coordinate>
                            <mileage>158044955</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ATY</key>
                                <value>DR</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601928488930-80</value>
                            </property>
                            <property>
                                <key>ALEN</key>
                                <value>86177</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48115.9</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6284</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T20:08:08.929</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>DRLEN</key>
                                <value>83000</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:17:39.240Z</time>
                            <coordinate>
                                <latitude>52.42518</latitude>
                                <longitude>4.80460</longitude>
                            </coordinate>
                            <mileage>158044955</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>55</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:18:03.664Z</time>
                            <coordinate>
                                <latitude>52.42519</latitude>
                                <longitude>4.80459</longitude>
                            </coordinate>
                            <mileage>158044955</mileage>
                            <heading>274</heading>
                            <speed>2</speed>
                            <property>
                                <key>STL</key>
                                <value>300000</value>
                            </property>
                            <property>
                                <key>STT</key>
                                <value>2020-10-05T20:11:31.230</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                            <property>
                                <key>STD</key>
                                <value>392434</value>
                            </property>
                        </traces>
                        <traces>
                            <type>167</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:19:04.902Z</time>
                            <coordinate>
                                <latitude>52.42442</latitude>
                                <longitude>4.80632</longitude>
                            </coordinate>
                            <mileage>158045165</mileage>
                            <heading>0</heading>
                            <speed>2</speed>
                            <property>
                                <key>FLP</key>
                                <value>99</value>
                            </property>
                            <property>
                                <key>FLC</key>
                                <value>10000</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>13</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:20:27.869Z</time>
                            <coordinate>
                                <latitude>52.42539</latitude>
                                <longitude>4.80761</longitude>
                            </coordinate>
                            <mileage>158045350</mileage>
                            <heading>10</heading>
                            <speed>30</speed>
                            <property>
                                <key>ATY</key>
                                <value>UN</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601928908423-79</value>
                            </property>
                            <property>
                                <key>ALEN</key>
                                <value>652755</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48116.2</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6285</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T20:09:35.106</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>DRLEN</key>
                                <value>143863</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>10</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:20:27.882Z</time>
                            <coordinate>
                                <latitude>52.42539</latitude>
                                <longitude>4.80761</longitude>
                            </coordinate>
                            <mileage>158045350</mileage>
                            <heading>10</heading>
                            <speed>30</speed>
                            <property>
                                <key>ATY</key>
                                <value>DR</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601929227861-78</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48116.2</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6286</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T20:20:27.861</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:21:20.857Z</time>
                            <coordinate>
                                <latitude>52.42598</latitude>
                                <longitude>4.80754</longitude>
                            </coordinate>
                            <mileage>158045445</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>13</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:21:20.798Z</time>
                            <coordinate>
                                <latitude>52.42598</latitude>
                                <longitude>4.80754</longitude>
                            </coordinate>
                            <mileage>158045445</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ATY</key>
                                <value>DR</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601929227861-78</value>
                            </property>
                            <property>
                                <key>ALEN</key>
                                <value>24122</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48116.2</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6286</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T20:20:27.861</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>DRLEN</key>
                                <value>19000</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>10</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:21:20.816Z</time>
                            <coordinate>
                                <latitude>52.42598</latitude>
                                <longitude>4.80754</longitude>
                            </coordinate>
                            <mileage>158045445</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ATY</key>
                                <value>UN</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601929280770-77</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48116.2</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6287</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T20:20:51.983</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:36:20.862Z</time>
                            <coordinate>
                                <latitude>52.42596</latitude>
                                <longitude>4.80756</longitude>
                            </coordinate>
                            <mileage>158045445</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.168</value>
                            </property>
                        </traces>
                        <traces>
                            <type>71</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:47:37.601Z</time>
                            <coordinate>
                                <latitude>52.42596</latitude>
                                <longitude>4.80755</longitude>
                            </coordinate>
                            <mileage>158045445</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.9160000000000004</value>
                            </property>
                        </traces>
                        <traces>
                            <type>13</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:48:58.938Z</time>
                            <coordinate>
                                <latitude>52.42764</latitude>
                                <longitude>4.80770</longitude>
                            </coordinate>
                            <mileage>158045660</mileage>
                            <heading>308</heading>
                            <speed>30</speed>
                            <property>
                                <key>ATY</key>
                                <value>UN</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601929280770-77</value>
                            </property>
                            <property>
                                <key>ALEN</key>
                                <value>1686944</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48116.4</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6287</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T20:20:51.983</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>DRLEN</key>
                                <value>63929</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.9160000000000004</value>
                            </property>
                        </traces>
                        <traces>
                            <type>10</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:48:58.956Z</time>
                            <coordinate>
                                <latitude>52.42764</latitude>
                                <longitude>4.80770</longitude>
                            </coordinate>
                            <mileage>158045660</mileage>
                            <heading>308</heading>
                            <speed>30</speed>
                            <property>
                                <key>ATY</key>
                                <value>DR</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601930938927-76</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48116.4</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6288</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T20:48:58.927</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.9160000000000004</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:53:58.976Z</time>
                            <coordinate>
                                <latitude>52.42670</latitude>
                                <longitude>4.84931</longitude>
                            </coordinate>
                            <mileage>158049435</mileage>
                            <heading>40</heading>
                            <speed>52</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:56:52.224Z</time>
                            <coordinate>
                                <latitude>52.42122</latitude>
                                <longitude>4.85375</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>167</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:56:52.219Z</time>
                            <coordinate>
                                <latitude>52.42122</latitude>
                                <longitude>4.85375</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLP</key>
                                <value>99</value>
                            </property>
                            <property>
                                <key>FLC</key>
                                <value>10000</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>13</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:56:52.157Z</time>
                            <coordinate>
                                <latitude>52.42122</latitude>
                                <longitude>4.85375</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ATY</key>
                                <value>DR</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601930938927-76</value>
                            </property>
                            <property>
                                <key>ALEN</key>
                                <value>445430</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48117.6</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6288</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T20:48:58.927</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>DRLEN</key>
                                <value>442000</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>10</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T20:56:52.175Z</time>
                            <coordinate>
                                <latitude>52.42122</latitude>
                                <longitude>4.85375</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ATY</key>
                                <value>UN</value>
                            </property>
                            <property>
                                <key>LID</key>
                                <value>500004f451601931412147-75</value>
                            </property>
                            <property>
                                <key>TFU</key>
                                <value>48117.6</value>
                            </property>
                            <property>
                                <key>SEQ</key>
                                <value>6289</value>
                            </property>
                            <property>
                                <key>SST</key>
                                <value></value>
                            </property>
                            <property>
                                <key>AST</key>
                                <value>2020-10-05T20:56:24.357</value>
                            </property>
                            <property>
                                <key>AID</key>
                                <value></value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T21:11:52.229Z</time>
                            <coordinate>
                                <latitude>52.42125</latitude>
                                <longitude>4.85369</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.132</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T21:26:52.236Z</time>
                            <coordinate>
                                <latitude>52.42126</latitude>
                                <longitude>4.85373</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>4.068</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T21:41:52.245Z</time>
                            <coordinate>
                                <latitude>52.42126</latitude>
                                <longitude>4.85372</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.564</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T21:56:52.253Z</time>
                            <coordinate>
                                <latitude>52.42124</latitude>
                                <longitude>4.85377</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.168</value>
                            </property>
                        </traces>
                        <traces>
                            <type>73</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-05T21:56:54.157Z</time>
                            <coordinate>
                                <latitude>52.42124</latitude>
                                <longitude>4.85377</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.168</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T05:14:21.599Z</time>
                            <coordinate>
                                <latitude>52.42126</latitude>
                                <longitude>4.85375</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.384</value>
                            </property>
                        </traces>
                        <traces>
                            <type>200</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T05:15:03.492Z</time>
                            <coordinate>
                                <latitude>52.42123</latitude>
                                <longitude>4.85376</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>GPRSSTATE</key>
                                <value>3</value>
                            </property>
                            <property>
                                <key>SIMID</key>
                                <value>8931440302015058585</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.204</value>
                            </property>
                            <property>
                                <key>SIMSL</key>
                                <value>0</value>
                            </property>
                            <property>
                                <key>GPRSNET</key>
                                <value>20404</value>
                            </property>
                            <property>
                                <key>DEVICE</key>
                                <value>500004f45</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T05:30:03.561Z</time>
                            <coordinate>
                                <latitude>52.42123</latitude>
                                <longitude>4.85372</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.8080000000000003</value>
                            </property>
                        </traces>
                        <traces>
                            <type>71</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T05:30:20.580Z</time>
                            <coordinate>
                                <latitude>52.42122</latitude>
                                <longitude>4.85372</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.456</value>
                            </property>
                        </traces>
                        <traces>
                            <type>7</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T05:30:47.612Z</time>
                            <coordinate>
                                <latitude>52.42121</latitude>
                                <longitude>4.85373</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.952</value>
                            </property>
                        </traces>
                        <traces>
                            <type>8</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T05:30:48.133Z</time>
                            <coordinate>
                                <latitude>52.42121</latitude>
                                <longitude>4.85373</longitude>
                            </coordinate>
                            <mileage>158051045</mileage>
                            <heading>0</heading>
                            <speed>2</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.952</value>
                            </property>
                        </traces>
                        <traces>
                            <type>167</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T05:32:20.832Z</time>
                            <coordinate>
                                <latitude>52.42154</latitude>
                                <longitude>4.85398</longitude>
                            </coordinate>
                            <mileage>158051280</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLP</key>
                                <value>99</value>
                            </property>
                            <property>
                                <key>FLC</key>
                                <value>500</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.3120000000000003</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T05:32:20.854Z</time>
                            <coordinate>
                                <latitude>52.42154</latitude>
                                <longitude>4.85398</longitude>
                            </coordinate>
                            <mileage>158051280</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.3120000000000003</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T05:47:20.863Z</time>
                            <coordinate>
                                <latitude>52.42154</latitude>
                                <longitude>4.85395</longitude>
                            </coordinate>
                            <mileage>158051280</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.672</value>
                            </property>
                        </traces>
                        <traces>
                            <type>71</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:00:00.123Z</time>
                            <coordinate>
                                <latitude>52.42154</latitude>
                                <longitude>4.85401</longitude>
                            </coordinate>
                            <mileage>158051280</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.8520000000000003</value>
                            </property>
                        </traces>
                        <traces>
                            <type>167</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:04:18.322Z</time>
                            <coordinate>
                                <latitude>52.42145</latitude>
                                <longitude>4.85403</longitude>
                            </coordinate>
                            <mileage>158051300</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLP</key>
                                <value>99</value>
                            </property>
                            <property>
                                <key>FLC</key>
                                <value>500</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.8800000000000003</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:04:18.333Z</time>
                            <coordinate>
                                <latitude>52.42145</latitude>
                                <longitude>4.85403</longitude>
                            </coordinate>
                            <mileage>158051300</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.8800000000000003</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:09:18.354Z</time>
                            <coordinate>
                                <latitude>52.42155</latitude>
                                <longitude>4.85389</longitude>
                            </coordinate>
                            <mileage>158051510</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.664</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:10:54.160Z</time>
                            <coordinate>
                                <latitude>52.42150</latitude>
                                <longitude>4.85395</longitude>
                            </coordinate>
                            <mileage>158051610</mileage>
                            <heading>0</heading>
                            <speed>4</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.628</value>
                            </property>
                        </traces>
                        <traces>
                            <type>71</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:10:50.300Z</time>
                            <coordinate>
                                <latitude>52.42151</latitude>
                                <longitude>4.85395</longitude>
                            </coordinate>
                            <mileage>158051510</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.664</value>
                            </property>
                        </traces>
                        <traces>
                            <type>71</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:10:54.270Z</time>
                            <coordinate>
                                <latitude>52.42150</latitude>
                                <longitude>4.85395</longitude>
                            </coordinate>
                            <mileage>158051610</mileage>
                            <heading>0</heading>
                            <speed>4</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.628</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:11:56.946Z</time>
                            <coordinate>
                                <latitude>52.42137</latitude>
                                <longitude>4.85396</longitude>
                            </coordinate>
                            <mileage>158051635</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.628</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:26:56.961Z</time>
                            <coordinate>
                                <latitude>52.42148</latitude>
                                <longitude>4.85393</longitude>
                            </coordinate>
                            <mileage>158051635</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.8160000000000003</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:30:44.221Z</time>
                            <coordinate>
                                <latitude>52.42127</latitude>
                                <longitude>4.85398</longitude>
                            </coordinate>
                            <mileage>158051640</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.42</value>
                            </property>
                        </traces>
                        <traces>
                            <type>71</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:30:41.473Z</time>
                            <coordinate>
                                <latitude>52.42127</latitude>
                                <longitude>4.85398</longitude>
                            </coordinate>
                            <mileage>158051635</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.42</value>
                            </property>
                        </traces>
                        <traces>
                            <type>71</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:30:44.313Z</time>
                            <coordinate>
                                <latitude>52.42127</latitude>
                                <longitude>4.85398</longitude>
                            </coordinate>
                            <mileage>158051640</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.42</value>
                            </property>
                        </traces>
                        <traces>
                            <type>167</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:33:14.683Z</time>
                            <coordinate>
                                <latitude>52.42128</latitude>
                                <longitude>4.85391</longitude>
                            </coordinate>
                            <mileage>158051640</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLP</key>
                                <value>99</value>
                            </property>
                            <property>
                                <key>FLC</key>
                                <value>500</value>
                            </property>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.096</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:33:14.691Z</time>
                            <coordinate>
                                <latitude>52.42128</latitude>
                                <longitude>4.85391</longitude>
                            </coordinate>
                            <mileage>158051640</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.096</value>
                            </property>
                        </traces>
                        <traces>
                            <type>71</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:36:00.182Z</time>
                            <coordinate>
                                <latitude>52.42135</latitude>
                                <longitude>4.85389</longitude>
                            </coordinate>
                            <mileage>158051640</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.096</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:41:00.203Z</time>
                            <coordinate>
                                <latitude>52.42122</latitude>
                                <longitude>4.85400</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.096</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:42:22.811Z</time>
                            <coordinate>
                                <latitude>52.42118</latitude>
                                <longitude>4.85404</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.3120000000000003</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (TAB)</source>
                            <time>2020-10-06T06:46:14.026Z</time>
                            <coordinate>
                                <latitude>52.49076</latitude>
                                <longitude>6.13909</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ACY</key>
                                <value>7.0</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T06:46:17.805Z</time>
                            <coordinate>
                                <latitude>52.49076</latitude>
                                <longitude>6.13909</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.096</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:57:22.819Z</time>
                            <coordinate>
                                <latitude>52.42128</latitude>
                                <longitude>4.85399</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.564</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:58:32.527Z</time>
                            <coordinate>
                                <latitude>52.42129</latitude>
                                <longitude>4.85405</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.988</value>
                            </property>
                        </traces>
                        <traces>
                            <type>71</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:58:32.332Z</time>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>0.0</value>
                            </property>
                            <property>
                                <key>LACY</key>
                                <value>2.988</value>
                            </property>
                            <property>
                                <key>LLAT</key>
                                <value>52.42130</value>
                            </property>
                            <property>
                                <key>LLON</key>
                                <value>4.85405</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:58:40.838Z</time>
                            <coordinate>
                                <latitude>52.42129</latitude>
                                <longitude>4.85407</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.988</value>
                            </property>
                        </traces>
                        <traces>
                            <type>72</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:59:13.537Z</time>
                            <coordinate>
                                <latitude>52.42129</latitude>
                                <longitude>4.85405</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.988</value>
                            </property>
                        </traces>
                        <traces>
                            <type>71</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T06:59:03.932Z</time>
                            <coordinate>
                                <latitude>52.42129</latitude>
                                <longitude>4.85406</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.988</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (TAB)</source>
                            <time>2020-10-06T07:01:15.150Z</time>
                            <coordinate>
                                <latitude>52.49072</latitude>
                                <longitude>6.13908</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ACY</key>
                                <value>7.0</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T07:01:17.822Z</time>
                            <coordinate>
                                <latitude>52.49071</latitude>
                                <longitude>6.13908</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.988</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T07:14:13.548Z</time>
                            <coordinate>
                                <latitude>52.42116</latitude>
                                <longitude>4.85397</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.7</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (TAB)</source>
                            <time>2020-10-06T07:16:16.012Z</time>
                            <coordinate>
                                <latitude>52.49071</latitude>
                                <longitude>6.13908</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ACY</key>
                                <value>6.0</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T07:16:17.830Z</time>
                            <coordinate>
                                <latitude>52.49071</latitude>
                                <longitude>6.13908</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.7</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T07:29:13.558Z</time>
                            <coordinate>
                                <latitude>52.42136</latitude>
                                <longitude>4.85389</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.06</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (TAB)</source>
                            <time>2020-10-06T07:31:16.085Z</time>
                            <coordinate>
                                <latitude>52.49073</latitude>
                                <longitude>6.13907</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ACY</key>
                                <value>7.0</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T07:31:17.837Z</time>
                            <coordinate>
                                <latitude>52.49073</latitude>
                                <longitude>6.13907</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.952</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T07:44:13.567Z</time>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>0.0</value>
                            </property>
                            <property>
                                <key>LACY</key>
                                <value>3.06</value>
                            </property>
                            <property>
                                <key>LLAT</key>
                                <value>52.42139</value>
                            </property>
                            <property>
                                <key>LLON</key>
                                <value>4.85387</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T07:44:13.972Z</time>
                            <coordinate>
                                <latitude>52.42140</latitude>
                                <longitude>4.85387</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.06</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T07:46:17.844Z</time>
                            <coordinate>
                                <latitude>52.49073</latitude>
                                <longitude>6.13909</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.096</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (TAB)</source>
                            <time>2020-10-06T07:50:57.985Z</time>
                            <coordinate>
                                <latitude>52.49074</latitude>
                                <longitude>6.13908</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ACY</key>
                                <value>7.0</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T07:59:13.981Z</time>
                            <coordinate>
                                <latitude>52.42123</latitude>
                                <longitude>4.85393</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.9160000000000004</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T08:01:17.867Z</time>
                            <coordinate>
                                <latitude>52.49074</latitude>
                                <longitude>6.13909</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.952</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (TAB)</source>
                            <time>2020-10-06T08:05:58.136Z</time>
                            <coordinate>
                                <latitude>52.49074</latitude>
                                <longitude>6.13907</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ACY</key>
                                <value>7.0</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T08:12:41.394Z</time>
                            <coordinate>
                                <latitude>52.49075</latitude>
                                <longitude>6.13907</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.24</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T08:14:13.989Z</time>
                            <coordinate>
                                <latitude>52.42133</latitude>
                                <longitude>4.85393</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.564</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T08:16:17.887Z</time>
                            <coordinate>
                                <latitude>52.49074</latitude>
                                <longitude>6.13907</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.24</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (TAB)</source>
                            <time>2020-10-06T08:21:00.093Z</time>
                            <coordinate>
                                <latitude>52.49073</latitude>
                                <longitude>6.13903</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ACY</key>
                                <value>8.0</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T08:29:13.997Z</time>
                            <coordinate>
                                <latitude>52.42119</latitude>
                                <longitude>4.85387</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.24</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T08:31:17.901Z</time>
                            <coordinate>
                                <latitude>52.49073</latitude>
                                <longitude>6.13903</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.24</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (TAB)</source>
                            <time>2020-10-06T08:36:00.987Z</time>
                            <coordinate>
                                <latitude>52.49073</latitude>
                                <longitude>6.13907</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ACY</key>
                                <value>8.0</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T08:44:14.006Z</time>
                            <coordinate>
                                <latitude>52.42117</latitude>
                                <longitude>4.85380</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.2760000000000002</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T08:46:17.909Z</time>
                            <coordinate>
                                <latitude>52.49074</latitude>
                                <longitude>6.13909</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.456</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (TAB)</source>
                            <time>2020-10-06T08:51:01.932Z</time>
                            <coordinate>
                                <latitude>52.49076</latitude>
                                <longitude>6.13908</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ACY</key>
                                <value>8.0</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T08:59:14.014Z</time>
                            <coordinate>
                                <latitude>52.42101</latitude>
                                <longitude>4.85384</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.42</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T09:01:17.916Z</time>
                            <coordinate>
                                <latitude>52.49076</latitude>
                                <longitude>6.13905</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>2.952</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (TAB)</source>
                            <time>2020-10-06T09:06:02.878Z</time>
                            <coordinate>
                                <latitude>52.49079</latitude>
                                <longitude>6.13907</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>ACY</key>
                                <value>7.0</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>43-BLK-9 (T4U)</source>
                            <time>2020-10-06T09:14:14.021Z</time>
                            <coordinate>
                                <latitude>52.42144</latitude>
                                <longitude>4.85404</longitude>
                            </coordinate>
                            <mileage>158051645</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>496.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.6</value>
                            </property>
                        </traces>
                        <traces>
                            <type>0</type>
                            <source>75-BPG-7 (T4U)</source>
                            <time>2020-10-06T09:16:17.933Z</time>
                            <coordinate>
                                <latitude>52.49076</latitude>
                                <longitude>6.13904</longitude>
                            </coordinate>
                            <mileage>69686170</mileage>
                            <heading>0</heading>
                            <speed>0</speed>
                            <property>
                                <key>FLA</key>
                                <value>9920.0</value>
                            </property>
                            <property>
                                <key>ACY</key>
                                <value>3.564</value>
                            </property>
                        </traces>
                        <mark>2020-10-12T12:04:10.640</mark>
                        <more>true</more>
                    </return>
                </ns2:pollTracesResponse>
            </env:Body>
        </env:Envelope>"""


        root = XMLElement.parse_string(sample_trimble_trace_data)

        response = root.children[1].children[0].find("return")

        self.assertTrue(response.has_child)

        first_tag = response.children[0]

        self.assertEqual(first_tag.full_tag, "traces")
        
        properties = first_tag.find_all("property")

        self.assertDictEqual(properties[0].to_dict(), {
            "tag": "property",
            "children": [
                {
                    "tag": "key",
                    "text": "FLA"
                },
                {
                    "tag": "value",
                    "text": "9920.0"
                }
            ]
        })        

        self.assertDictEqual(properties[2].to_dict(), {
            "tag": "property",
            "children": [
                {
                    "tag": "key",
                    "text": "AFRE"
                },
                {
                    "tag": "value",
                    "children": [
                        {
                            "tag": "Report",
                            "attributes": {"id": "WAITING", "version": "2", "timestamp": "1455703892236"},
                            "children": [
                                {
                                    "tag": "Answer",
                                    "attributes": {"question": "REASON"},
                                    "children": [
                                        {
                                            "tag": "Value",
                                            "attributes": {"option": "02", "value": "true"}
                                        }
                                    ]
                                }
                            ]
                        }
                    ]
                }
            ]
        })

        # print(response.find('traces').text)


class EnvChoices:
    NORMAL = "normal"
    E2E = 'e2e'
    TEST = 'test'
    
class ResponseChoices:
    success = "success"
    already_exists = "already exists"
    does_not_exist = "does not exist"
    same_object = "same object"
    not_logged = "not logged"
    not_allowed = "not allowed"
    different_type = "different type"
    not_balanced = "not balanced"
    failure = "failure"
    not_valid = "not valid"
    limit_exceeded = "limit exceeded"


# Python
import json

# Django
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views import View as DjangoView
from django.http import HttpRequest

class CSRFExemptMixin(object):
    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super(CSRFExemptMixin, self).dispatch(*args, **kwargs) # type: ignore
    
class View(DjangoView):

    def dispatch(self, request: HttpRequest, *args, **kwargs):
        if request.method == "PUT":
            request.PUT = json.loads(request.body.decode('utf-8').replace("'", '"')) # type: ignore
        elif request.method == "DELETE":
            request.DELETE = json.loads(request.body.decode('utf-8').replace("'", '"')) # type: ignore
        return super().dispatch(request, *args, **kwargs)


# Python
from typing import Any, TypeVar, Callable

# Django
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.db.models.manager import BaseManager


_Z = TypeVar("_Z")


def paginate(entries: BaseManager[Any], page_number: int, limit: int, sort: str | None, serializer: Callable, *args, **kwargs) -> dict:
    
    paginator = Paginator(entries, limit)

    try:
        page = paginator.page(page_number)
    except PageNotAnInteger:
        page = paginator.page(1)
    except EmptyPage:
        page = paginator.page(paginator.num_pages)

    return {
        "data": serializer(page.object_list, *args, **kwargs),
        "limit": limit,
        "page": page_number,
        "startIndex": page.start_index(),
        "hasNext": page.has_next(),
        "sort": sort,
        "total": paginator.count,
        "totalPage": paginator.num_pages,
    }


def paginate_without_serialization(entries: BaseManager[_Z], page_number: int, limit: int, sort: str | None) -> tuple[BaseManager[_Z], dict]: # type: ignore
    paginator = Paginator(entries, limit)

    try:
        page = paginator.page(page_number)
    except PageNotAnInteger:
        page = paginator.page(1)
    except EmptyPage:
        page = paginator.page(paginator.num_pages)

    return page.object_list, { # type: ignore
        "limit": limit,
        "page": page_number,
        "startIndex": page.start_index(),
        "hasNext": page.has_next(),
        "sort": sort,
        "total": paginator.count,
        "totalPage": paginator.num_pages,
    }


class InputError(Exception):

    def __init__(self, obj, message=None):
        self.obj = obj
        self.message = message
        

class Failure(Exception):

    def __init__(self, reason: str):
        self.reason = reason

class AlreadyExist(Exception):

    def __init__(self, obj: str, reason: str | None = None):
        self.obj = obj
        self.reason = reason

class NotAllowed(Exception):

    def __init__(self, reason):
        self.reason = reason



# Python
from datetime import datetime, timedelta

# Django
from django.test import TestCase

from django_carpet.validators import date_time_validation, string_validation, choice_validation, boolean_validation, integer_validation
from django_carpet.exceptions import InputError


class ValidaotrsTest(TestCase):

    def test_date_time_validators(self):
        now = datetime.now()
        valid_iso = now.isoformat()
        min_iso = (datetime.now() + timedelta(days=10)).isoformat()

        # Testing the date time validation when the value is None
        try:
            date_time_validation(None, "dateTime")
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'dateTime')
        except Exception as e:
            self.fail(e)


        # Testing the date time validation when the value is not an ISO format date time
        try:
            date_time_validation('abc', "dateTime")
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'dateTime')
        except Exception as e:
            self.fail(e)

        # Testing the date time validation when the value is a valid but before min date
        try:
            date_time_validation(valid_iso, "dateTime", False, min_iso)
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'dateTime')
        except Exception as e:
            self.fail(e)

        # Testing the date time validation when the value is None and empty is allowed
        self.assertIsNone(date_time_validation(None, 'dateTime', True))

        # Testing the date time validation when the value is an empty string and empty is allowed
        self.assertIsNone(date_time_validation("", 'dateTime', True))
        
        # Testing the date time validation when the value is valid without min date
        self.assertEqual(
            date_time_validation(valid_iso, 'dateTime'),
            now
        )

        # Testing the date time validation when the value is valid with a min date
        self.assertEqual(
            date_time_validation(valid_iso, 'dateTime', False, (datetime.now() - timedelta(days=10))),
            now
        )
        

    def test_string_validators(self):

        # Testing the string validation when the value is None
        try:
            string_validation(None, 10, 'string')
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'string')
        except Exception as e:
            self.fail(e)


        # Testing the string validation when the value is empty string
        try:
            string_validation("", 10, 'string')
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'string')
        except Exception as e:
            self.fail(e)


        # Testing the string validation when the text is too long
        try:
            string_validation('1234567891011', 10, "string")
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'string')
        except Exception as e:
            self.fail(e)

        self.assertEqual(type(string_validation("abc", 10, 'string')), str)
        self.assertEqual(type(string_validation("abc", 10, 'string', empty=True)), str)
        self.assertIsNone(string_validation(None, 10, 'string', empty=True))
        self.assertIsNone(string_validation("", 10, 'string', empty=True))
    

    def test_choice_validators(self):

        # Testing the choice validation when the value is None
        try:
            choice_validation(None, ["one", "two"], "choice")
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'choice')
        except Exception as e:
            self.fail(e)


        # Testing the choice validation when the value is empty string
        try:
            choice_validation("", ["one", "two"], 'choice')
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'choice')
        except Exception as e:
            self.fail(e)


        # Testing the choice validation when the value is not among the options
        try:
            choice_validation('three', ["one", "two"], "choice")
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'choice')
        except Exception as e:
            self.fail(e)

        self.assertEqual(type(choice_validation("one", ["one", "two"], 'choice')), str)
        self.assertEqual(type(choice_validation("one", ["one", "two"], 'choice', True)), str)
        self.assertIsNone(choice_validation(None, ["one", "two"], 'choice', null=True))
        self.assertIsNone(choice_validation("", ["one", "two"], 'choice', null=True))


    def test_boolean_validators(self):

        # Testing the boolean validation if the value is None
        try:
            boolean_validation(None, 'bool')
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'bool')
        except Exception as e:
            self.fail(e)

        # Testing the boolean validation if the value is an invalid string
        try:
            boolean_validation('other', 'bool')
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'bool')
        except Exception as e:
            self.fail(e)
            
        self.assertTrue(boolean_validation(True, 'bool'))
        self.assertTrue(boolean_validation('true', 'bool'))
        self.assertFalse(boolean_validation(False, 'bool'))
        self.assertFalse(boolean_validation('false', 'bool'))

    def test_integer_validators(self):

        # Testing the integer validation when the value is None and empty is false
        try:
            integer_validation(None, 0, 100, 'number', empty=False)
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'number')
        except Exception as e:
            self.fail(e)

        try:
            integer_validation('', 0, 100, 'number', empty=False)
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'number')
        except Exception as e:
            self.fail(e)
            
        # Testing the integer validation when the value is None and empty is True
        self.assertIsNone(integer_validation(None, 0, 100, 'number', empty=True))
        self.assertIsNone(integer_validation("", 0, 100, 'number', empty=True))
        
        # Passing an invalid text as number
        try:
            integer_validation('ss', 0, 100, 'number', empty=False)
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'number')
        except Exception as e:
            self.fail(e)

        # Passing a number smaller than the minimum
        try:
            integer_validation('0', 1, 100, 'number', empty=False)
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'number')
        except Exception as e:
            self.fail(e)

        try:
            integer_validation(0, 1, 100, 'number', empty=False)
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'number')
        except Exception as e:
            self.fail(e)

        # Passing a number bigger than the minimum
        try:
            integer_validation('200', 0, 100, 'number', empty=False)
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'number')
        except Exception as e:
            self.fail(e)

        try:
            integer_validation(200, 0, 100, 'number', empty=False)
            self.fail("exception was not raised")
        except InputError as e:
            self.assertEqual(e.obj, 'number')
        except Exception as e:
            self.fail(e)

        # Passing valid values 
        self.assertEqual(integer_validation(200, 0, 300, 'number', empty=False), 200)
        self.assertEqual(integer_validation('200', 0, 300, 'number', empty=False), 200)


"""
Sample settings file

Generated by 'django-admin startproject' using Django 4.0.

For more information on this file, see
https://docs.djangoproject.com/en/2.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.0/ref/settings/
"""

import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

PROJECT_VERSION = "0.0.1"


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "test secret"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

ALLOWED_HOSTS = []
CSRF_TRUSTED_ORIGINS=[]
LOGIN_URL = '/login/'

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.sites',
    'django.contrib.sitemaps',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    # 'django.contrib.sessions.middleware.SessionMiddleware',
    'foundation.middleware.CustomSessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'foundation.middleware.CustomCsrfMiddleware',
    # 'django.contrib.auth.middleware.AuthenticationMiddleware',
    'foundation.middleware.CustomAuthenticationMiddleware',
    # 'django.contrib.messages.middleware.MessageMiddleware',
    'foundation.middleware.CustomMessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'foundation.middleware.RemoteAuthenticationMiddleware',
    'foundation.middleware.ServerErrorReporterMiddleware'
]


TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, 'templates')
            ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]


TEMPLATE_DIRS = (
    # Put strings here, like "/home/html/django_templates" or "C:/www/django/templates".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
    os.path.join(BASE_DIR, 'templates').replace('\\', '/'),
)

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
    # 'django.template.loaders.eggs.Loader',
)

WSGI_APPLICATION = 'test.wsgi.application'


# Database
# https://docs.djangoproject.com/en/2.0/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db_test.sqlite3'),
    }
}

# Password validation
# https://docs.djangoproject.com/en/2.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.0/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_TZ = True


SECURE_CONTENT_TYPE_NOSNIFF = True
SECURE_BROWSER_XSS_FILTER = True
SECURE_SSL_REDIRECT = False
SESSION_COOKIE_SECURE = False
CSRF_COOKIE_SECURE = False


# if os.environ.get('DJANGO_DEVELOPMENT') is not None:
#     from .settings_dev import *  # noqa: F403